/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Author: Charles Kerr <charles@rebelbase.com>
 *
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <pan/base/debug.h>
#include <pan/base/message-identifier.h>
#include <pan/base/pan-glib-extensions.h>

/**
***  GObject stuff
**/

static GObjectClass * parent_class = NULL;

static void
pan_message_identifier_init (MessageIdentifier * mid, MessageIdentifierClass * klass)
{
	mid->line_qty = 0ul;
	mid->message_id = NULL;
	mid->message_sources = g_ptr_array_new ();
}

static void
pan_message_identifier_finalize (GObject * object)
{
	int i;
	MessageIdentifier * mid = (MessageIdentifier*) object;

	for (i=0; i<mid->message_sources->len; ++i)
	{
		MessageSource * source = (MessageSource*) g_ptr_array_index (mid->message_sources, i);
		g_free (source->server_name);
		g_free (source->group_name);
		g_free (source);
	}
	g_ptr_array_free (mid->message_sources, TRUE);
	g_free (mid->message_id);

	mid->message_sources = NULL;
	mid->message_id = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
pan_message_identifier_class_init (MessageIdentifierClass * klass)
{
	GObjectClass * object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_ref (G_TYPE_OBJECT);

	object_class->finalize = pan_message_identifier_finalize;
}

GType
pan_message_identifier_get_type (void)
{
	static GType type = 0;

	if (!type)
	{
		static const GTypeInfo info =
		{
			sizeof (MessageIdentifierClass),
			NULL, /* base_class_init */
			NULL, /* base_class_finalize */
			(GClassInitFunc) pan_message_identifier_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (MessageIdentifier),
			16,   /* n_preallocs */
			(GInstanceInitFunc) pan_message_identifier_init
		};

		type = g_type_register_static (G_TYPE_OBJECT, "MessageIdentifier", &info, 0);
	}

	return type;
}

/**
***  MessageIdentifier Stuff
**/

MessageIdentifier*
message_identifier_new (const char * message_id)
{
	MessageIdentifier * mid;

	mid = g_object_new (PAN_TYPE_MESSAGE_IDENTIFIER, NULL, NULL);
	mid->message_id = g_strdup (message_id);

	return mid;
}

MessageIdentifier*
message_identifier_new_from_article (const Article * article)
{
	MessageIdentifier * mid;

	g_return_val_if_fail (article_is_valid(article), NULL);

	mid = message_identifier_new (article_get_message_id(article));
	message_identifier_add_source_from_article (mid, article);

	return mid;
}

void
message_identifier_add_source (MessageIdentifier  * mid,
	                       const char         * server,
	                       const char         * group,
	                       gulong               number)
{
	MessageSource * source;

	g_return_if_fail (PAN_IS_MESSAGE_IDENTIFIER(mid));
       
	source = g_new (MessageSource, 1);
	source->server_name = g_strdup (server);
	source->group_name = g_strdup (group);
	source->number = number;
	g_ptr_array_add (mid->message_sources, source);
}

static void
add_source_from_article_xreffunc (Server * server, Group * group, gulong number, gpointer user_data)
{
	MessageIdentifier * mid = MESSAGE_IDENTIFIER(user_data);

	message_identifier_add_source (mid, server->name, group->name, number);
}

void
message_identifier_add_source_from_article  (MessageIdentifier  * mid,
                                             const Article      * article)
{
	g_return_if_fail (PAN_IS_MESSAGE_IDENTIFIER(mid));
	g_return_if_fail (article_is_valid(article));

	mid->line_qty = article->linecount;
	message_identifier_add_source (mid, article->group->server->name, article->group->name, article->number);
	article_xref_foreach (article, add_source_from_article_xreffunc, mid, SERVER_GROUPS_ALL, FALSE);
}

gboolean
message_identifier_is_valid (MessageIdentifier * mid)
{
	g_return_val_if_fail (PAN_IS_MESSAGE_IDENTIFIER(mid), FALSE);
	g_return_val_if_fail (is_nonempty_string(mid->message_id), FALSE);

	return TRUE;
}

MessageSource*
message_identifier_get_source_for_server   (MessageIdentifier  * mid,
                                            const char         * server)
{
	int i;
	MessageSource * retval = NULL;

	/* sanity clause */
	g_return_val_if_fail (PAN_IS_MESSAGE_IDENTIFIER(mid), NULL);
	g_return_val_if_fail (is_nonempty_string(server), NULL);

	/* return the first for the matching that server */
	for (i=0; !retval && i<mid->message_sources->len; ++i) {
		MessageSource * source = (MessageSource*) g_ptr_array_index (mid->message_sources, i);
		if (!pan_strcmp (server, source->server_name))
			retval = source;
	}

	return retval;
}

const char*
message_identifier_get_primary_group (MessageIdentifier * mid)
{
	const char * retval = NULL;

	/* sanity clause */
	g_return_val_if_fail (PAN_IS_MESSAGE_IDENTIFIER(mid), NULL);

	/* find the first group name */
	if (mid->message_sources->len) {
		MessageSource * source = (MessageSource*) g_ptr_array_index (mid->message_sources, 0);
		retval = source->group_name;
	}

	return retval;
}
