/*
 * Pan - A Newsreader for X
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <pan/base/acache.h>
#include <pan/base/article.h>
#include <pan/base/base-prefs.h>
#include <pan/base/debug.h>
#include <pan/base/gnksa.h>
#include <pan/base/log.h>
#include <pan/base/message-check.h>
#include <pan/base/serverlist.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/pan-i18n.h>
#include <pan/base/util-mime.h>

#include <pan/globals.h>
#include <pan/message-check-ui.h>
#include <pan/nntp.h>
#include <pan/queue.h>
#include <pan/smtp.h>
#include <pan/task-post.h>

/*********************
**********************  Defines / Enumerated types
*********************/

enum MessageSendResult
{
	SMTP_FAILED     = (1<<0),
	NNTP_FAILED     = (1<<1)
};

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PRIVATE ROUTINES
************/

/**
 * Populate a hashtable of internal headers in a message
 */
static void
build_nntp_message_foreach_build (const char * name, const char * value, gpointer user_data) {
	if (article_header_is_internal (name))
		g_hash_table_insert ((GHashTable*)user_data, g_strdup(name), g_strdup(value));
}

/**
 * Removes the specified header from the message
 */
static void
build_nntp_message_foreach_remove (gpointer key, gpointer value, gpointer header) {
	g_mime_header_remove ((GMimeHeader*)header, (const char*)key);
}

/**
 * Adds the specified header to the message
 */
static void
build_nntp_message_foreach_add (gpointer key, gpointer value, gpointer header) {
	g_mime_header_add ((GMimeHeader*)header, (const char*)key, (const char*)value);
}

/**
 * Build the message that will be passed to the news server.
 * Error-checking is done elsewhere; by the time we reach this
 * point we assume it's in the clear and only work on building
 * the message.
 */
char*
build_nntp_message (GMimeMessage * message)
{
	char * tmp;
	GHashTable * internal_headers;
	GMimeHeader * headers;
	debug_enter ("build_nntp_message");

	/* sanity checks... */
	g_return_val_if_fail (GMIME_IS_MESSAGE(message), NULL);

	/* update the message's date -- it may have been in sendlater for awhile */
	if (1) {
		time_t now = time (0);
		const int offset_secs = tzoffset_sec (&now);
		g_mime_message_set_date (message, now, (offset_secs/3600)*100);
	}

	/* temporarily remove the internal headers so that they don't show up in _to_string() */
	headers = GMIME_OBJECT(message)->headers;
	internal_headers = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
	g_mime_header_foreach (headers, build_nntp_message_foreach_build, internal_headers);
	g_hash_table_foreach (internal_headers, build_nntp_message_foreach_remove, headers);

	/* make sure each line ends in \r\n */
	tmp = g_mime_message_to_string (message);
	replace_gstr (&tmp, pan_substitute (tmp, "\n.", "\n.."));
	replace_gstr (&tmp, pan_substitute (tmp, "\n", "\r\n"));
	replace_gstr (&tmp, pan_substitute (tmp, "\r\r\n.", "\r\n"));

	/* add back the internal headers that we removed earlier */
	g_hash_table_foreach (internal_headers, build_nntp_message_foreach_add, headers);

	/* cleanup */
	g_hash_table_destroy (internal_headers);
	debug_exit ("build_nntp_message");
	return tmp;
}

/**
 * At this point we've already gotten a task and socket from the queue
 * and have done all the error-checking with the headers/body.  We're
 * ready to post the message and/or send the mail.
 */
static int
send_message_now (StatusItem           * item,
                  PanSocket            * sock,
                  Server               * server,
                  GMimeMessage         * message)
{
	gint retval = 0;
	const char * subject;
	const char * mail_to;
	const char * post_to;
	debug_enter ("send_message_now");

	/* initialize variables */
       	subject = g_mime_message_get_subject (message);
       	mail_to = g_mime_message_get_header (message, PAN_TO);
       	post_to = g_mime_message_get_header (message, PAN_NEWSGROUPS);

	/* mail the message */
	if (is_nonempty_string(mail_to))
	{
		char * leader = NULL;
		int val;

		if (is_nonempty_string(post_to))
		{
			leader = g_strdup_printf (
			_("[This is an email copy of a Usenet post to \"%s\"]"),
			post_to);
		}

		val = smtp_send (item, message, leader);

		/* mail the message */
		if (!val)
		{
			log_add_va (LOG_INFO, _("Email \"%s\" sent."), subject);
			g_mime_object_remove_header (GMIME_OBJECT(message), PAN_TO);
		}
		else
		{
			log_add_va (LOG_ERROR, _("Email \"%s\" not sent."), subject);
			retval |= SMTP_FAILED;
		}

		/* cleanup */
		g_free (leader);
	}

	/* post the message */
	if (is_nonempty_string(post_to))
	{
		char * pch;
		char * message_id = NULL;

		/* build the message.
		 * strip out the message-id if necessary. */
		if (!server->gen_msgid) {
			message_id = g_strdup (g_mime_message_get_message_id (message));
			g_mime_message_set_message_id (message, NULL);
		}
		pch = build_nntp_message (message);
		if (message_id != NULL) {
			g_mime_message_set_message_id (message, message_id);
			replace_gstr (&message_id, NULL);
		}

		/* let the user know what we're doing */
		status_item_emit_status_va (STATUS_ITEM(item), _("Posting article \"%s\""), subject);

		/* post the part */
		if (nntp_post(item,sock,pch) == TASK_SUCCESS) {
			log_add_va (LOG_INFO, _("Article \"%s\" posted."), subject);
			g_mime_object_remove_header (GMIME_OBJECT(message), PAN_NEWSGROUPS);
		} else {
			log_add_va (LOG_ERROR, _("Article \"%s\" not posted."), subject);
			retval |= NNTP_FAILED;
		}
		
		/* cleanup */
		g_free (pch);
	}


	debug_exit ("send_message_now");
	return retval;
}

static void
task_post_destructor (PanObject * obj)
{
	debug_enter ("task_post_destructor");

	/* sanity clause */
	g_return_if_fail (obj!=NULL);

	/* task-post dtor */
	g_object_unref (TASK_POST(obj)->message);

	/* destroy parent class */
	task_destructor (obj);

	debug_exit ("task_post_destructor");
}

static char*
task_post_describe (const StatusItem* si)
{
	GMimeMessage * message = TASK_POST(si)->message;
	const gboolean posting = g_mime_message_get_header (message, PAN_NEWSGROUPS) != NULL;
	const gboolean mailing = g_mime_message_get_header (message, PAN_TO) != NULL;
	const char * fmt;

	if (posting && mailing)
		fmt = _("Posting and mailing article \"%s\"");
	else if (posting)
		fmt = _("Posting article \"%s\"");
	else if (mailing)
		fmt = _("Mailing article \"%s\"");
	else 
		fmt = "BUG!";

	return g_strdup_printf (fmt, g_mime_message_get_subject(message));
}

static int
task_post_run (Task* task)
{
	int status;
	debug_enter ("task_post_run");

	/* try to send the message */
	status = send_message_now (STATUS_ITEM(task),
	                           TASK(task)->sock,
	                           TASK(task)->server,
	                           TASK_POST(task)->message);

	if (status != 0)
	{
		GMimeMessage * message;
		GString * text;

		/* write the message back to pan.sendlater, because
		 * PAN_TO or PAN_NEWSGROUPS may have been cleared */
	       	message = TASK_POST(task)->message;
		folder_add_message (serverlist_get_named_folder(PAN_SENDLATER), message);

	       	text = g_string_new (NULL);
		if (status & SMTP_FAILED)
			g_string_append (text, _("Email send failed.\n"));
		if (status & NNTP_FAILED)
			g_string_append (text, _("Usenet posting failed.\n"));
		if (text->len != 0) {
			g_string_append (text, _("Check Tools|Log Viewer for more information.\n"));
			g_string_append (text, _("Your message was saved in the folder \"pan.sendlater\""));
			log_add (LOG_ERROR|LOG_URGENT, text->str);
		}

		g_string_free (text, TRUE);
	}
	else /* successful; dispose of the article object. */
	{
		GMimeMessage * message = TASK_POST(task)->message;
		folder_remove_message (serverlist_get_named_folder(PAN_SENDLATER), message);
		folder_add_message (serverlist_get_named_folder(PAN_SENT), message);
	}

	/* if we failed, we make it hopeless because we've saved
	 * the message in pan.sendlater anyway.  This way the user
	 * can try to correct the problem before resending.
	 */
	debug_exit ("task_post_run");
	return status==0 ? TASK_SUCCESS : TASK_FAIL_HOPELESS;
}

/************
*************  PUBLIC ROUTINES
************/

PanObject*
task_post_new (Server * server, GMimeMessage * message)
{
	TaskPost * post = NULL;
	debug_enter ("task_post_new");

	/* sanity clause */
	g_return_val_if_fail (server_is_valid(server), NULL);
	g_return_val_if_fail (GMIME_IS_MESSAGE(message), NULL);

	/* create the object */
       	post = g_new0 (TaskPost, 1);
        debug1 (DEBUG_PAN_OBJECT, "task_post_new: %p", post);

	/* initialize the parent class */
	task_constructor (TASK(post), task_post_destructor,
	                              task_post_describe,
	                              task_post_run,
	                              server, TRUE, TRUE);
	TASK(post)->type = TASK_TYPE_POST;

	/* initialize the task-post */
	post->message = message;
	g_object_ref (message);

	debug_exit ("task_post_new");
	return PAN_OBJECT(post);
}

/*****
******
*****/


/****
*****
*****   UTILITIES
*****
****/

static void
message_checked_cb (Server        * server,
                    GMimeMessage  * message,
                    GoodnessLevel   goodness,
                    gpointer        user_data)
{
	Task * task = NULL;

	if (goodness == OKAY)
		task = TASK (task_post_new (server, message));
	if (task != NULL)
		queue_add (task);
}

void
queue_message_for_posting (Server          * server,
                           GMimeMessage    * message,
                           gboolean          message_check_first)
{
	/* sanity check */
	g_return_if_fail (server_is_valid (server));
	g_return_if_fail (GMIME_IS_MESSAGE (message));

	/* if desired, make sure the message is okay. */
	if (message_check_first)
		message_check_and_prompt (Pan.window, server, message, message_checked_cb, NULL);
	else
		message_checked_cb (server, message, OKAY, NULL);
}

static void
flush_sendlater_foreach (GMimeMessage * message, gpointer user_data)
{
	/* find the server that we're posting with */
	Server * server = NULL;
	const char * server_name = g_mime_message_get_header (message, PAN_SERVER);
	if (server_name != NULL)
		server = serverlist_get_named_server (server_name);
	if (server == NULL)
		server = serverlist_get_active_server ();

	/* queue the message */
	queue_message_for_posting (server, message, TRUE);
}

void
flush_sendlater_messages (void)
{
	acache_path_foreach (PAN_SENDLATER, flush_sendlater_foreach, NULL);
}
